import requests
import typing
import numpy
import json
from datetime import datetime, date
from dateutil import parser

from dacite import from_dict
from helmet.utils.types import Run, Explanation, AlternativesExplanation, ContrastiveExplanation, SaliencyExplanation, explanation_name_to_class
from dataclasses import asdict

numbers: tuple = tuple([numpy.int_, numpy.intc, numpy.intp, numpy.int8])
floats: tuple =  tuple([numpy.float_, numpy.float16, numpy.float32, numpy.float64])

class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, numbers):
            return int(obj)
        elif isinstance(obj, floats):
            return float(obj)
        elif isinstance(obj,(numpy.ndarray,)):
            return obj.tolist()
        elif isinstance(obj, Explanation):
            return asdict(obj)
        return json.JSONEncoder.default(self, obj)

def serialize(obj) -> dict:
    """ Serialize the object to a dictionary """
    return json.loads(json.dumps(obj, cls=NumpyEncoder))

def update_app(url: str, route: str, body: dict[str, typing.Any]):
    """ Update the app with the new model and tokenizer. 
    args: 
    url: str: the url of the app
    route: str: the route to update the app (format: /model or /tokenizer) 
    body: dict: the body of the request """
    
    if url is None or route is None:
        raise ValueError(f"url cannot be None url: {url} route:{route}")
    try :
        r = requests.post(f"{url}{route}", json=serialize(body))
        r.raise_for_status()
        print("updated app, result: ", r.json())
        
    except Exception as e:
        print(e)
        raise ValueError(f"Failed to get app. Is it running? url: {url} route: {route}")
    

def get_run(url: str, run_id: str) -> Run | None:
    """ Get the run from the platform """
    if url is None or run_id is None:
        raise ValueError(f"url cannot be None url: {url} run_id:{run_id}")
    final_url = f"{url}/runs/{run_id}"

    try:
        r = requests.get(final_url)
        r.raise_for_status()
    except Exception as e:
        print(e)
        raise ValueError(f"Failed to get run. Is the platform running? url: {final_url}")
    
    try:
        d = r.json()
        d["date"] = parser.parse(d["date"])
        # form = "%Y-%m-%dT%H:%M:%S.%fZ"
        # d["date"] = datetime.strptime(d["date"], form)

        expls = []
        for exp in d["explanations"]:
            expl_method = exp.pop("explanation_method")
            expl_class = explanation_name_to_class[expl_method]
            expls.append(expl_class(**exp))

        d["explanations"] = expls
        return from_dict(data_class=Run, data=d)

    except Exception as e:
        print(e)
        return None
    
def get_or_create_project(url: str, project_name: str, task: str) -> str:
    """ Get or create the project """
    if url is None or project_name is None or task is None:
        raise ValueError(f"url cannot be None url: {url} project_name:{project_name} description:{task}")
    final_url = f"{url}/project"

    try:
        r = requests.get(final_url)
        r.raise_for_status()
    except Exception as e:
        print(e)
        raise ValueError(f"Failed to get projects. Is the platform running? url: {final_url}")
    
    projects = r.json()
    for project in projects:
        if project["projectName"] == project_name:
            print("found project with the same name. Using the same project.")
            return project["_id"].__str__()
    

    print("creating new project")
    r = requests.post(final_url, json={"projectName": project_name, "task": task})
    if r.status_code != 200:
        raise ValueError(f"Failed to create project. Status code: {r.status_code}")
    
    return r.json()["_id"]


import torch
import numpy as np
from captum.attr import (
    InputXGradient, Saliency, LayerIntegratedGradients, 
    LLMGradientAttribution, TextTokenInput, LLMAttribution, IntegratedGradients
)
from typing import Dict, Any
from functools import partial
from torch.nn import functional as F

def compute_gradient(wrapper, prompt, input, output, gradient_type):
    # TODO: add with torch.no_grad():?

    def model_forward(inp, model, extra_forward_args: Dict[str, Any] = {}):
        output = model(input_ids=inp, **extra_forward_args)
        return F.softmax(output.logits, dim=1)
    
    input_embeds = wrapper.get_input_embeddings(prompt)
    attention_mask = input["attention_mask"]

    forward_func = partial(model_forward, model=wrapper.model, extra_forward_args={"attention_mask": attention_mask})

    if gradient_type == "input_x_gradient":
        # According to Luo and Specia (2024), inputxgradient has a lot of overhead because they also compute the gradients
        # of the input based on a reference input.
        lig = InputXGradient(forward_func)
    else:
        lig = Saliency(forward_func)
    attributions = lig.attribute(inputs=input_embeds, target=output)
    attributions = attributions.detach().cpu().numpy()
    r = attributions[0, :, :]
    attr = r.sum(axis=1)
    attr = wrapper.normalize(attr)

    return attr

# Adapted from Interpret-LM
# Adapted from AllenNLP Interpret and Han et al. 2020
def register_embedding_list_hook(model, embedding_layer, embeddings_list):
    def forward_hook(module, inputs, output):
        embeddings_list.append(output.squeeze(0).clone().cpu().detach().numpy())
    handle = embedding_layer.register_forward_hook(forward_hook)
    return handle

def register_embedding_gradient_hooks(model, embedding_layer, embeddings_gradients):
    def hook_layers(module, grad_in, grad_out):
        embeddings_gradients.append(grad_out[0].detach().cpu().numpy())
    hook = embedding_layer.register_full_backward_hook(hook_layers)
    return hook

def analyze_token(wrapper, input_ids, input_mask, batch=0, correct=None, foil=None):
    # Get model gradients and input embeddings
    torch.enable_grad()
    model = wrapper.model
    embedding_layer = wrapper.embeddings

    model.eval()
    embeddings_list = []
    handle = register_embedding_list_hook(model, embedding_layer, embeddings_list)
    embeddings_gradients = []
    hook = register_embedding_gradient_hooks(model, embedding_layer, embeddings_gradients)
    if correct is None:
        correct = input_ids[-1]
        input_ids = input_ids[:-1]
        input_mask = input_mask[:-1]

    input_ids_new = torch.tensor(input_ids.clone().detach())
    input_ids_unsqueezed = input_ids_new.unsqueeze(0)

    # A = model.generate(input_ids=input_ids_unsqueezed, attention_mask=input_mask)
    A = model(input_ids=input_ids_unsqueezed, output_attentions=False)

    if foil is not None and correct != foil:
        (A.logits[0][-1][correct]-A.logits[0][-1][foil]).backward()
    else:
        # Take the last logits and backpropagate the gradient
        p = A.logits[0][-1][correct]
        p.backward()
    handle.remove()
    hook.remove()

    import numpy as np
    return np.array(embeddings_gradients).squeeze(), np.array(embeddings_list).squeeze()

def input_x_gradient(grads, embds, normalize=False):
    input_grad = np.sum(grads * embds, axis=-1).squeeze()

    if normalize:
        norm = np.linalg.norm(input_grad, ord=1)
        input_grad /= norm
        
    return input_grad

# def analyze_token(wrapper, prompt, target):
#     ig = Saliency(forward_func=wrapper.model)

#     # LLM attribution
#     llm_attr = LLMAttribution(ig, wrapper.tokenizer)
    
#     # TextTokenInput is the only supported input for LLMGradientAttribution
#     input = TextTokenInput(
#         prompt, 
#         wrapper.tokenizer,
#         skip_tokens=[1],  # skip the special token for the start of the text <s>
#     )
    
#     res = llm_attr.attribute(input, target=output)
#     return res.seq_attr.detach().cpu().numpy()


#     # # Define the attribution method
#     # lig = Saliency(forward_func)
#     # attributions = lig.attribute(inputs=input_embeds, target=target)
#     # attributions = attributions.detach().cpu().numpy()
#     # r = attributions[0, :, :]
#     # attr = r.sum(axis=1)
#     # attr = wrapper.normalize(attr)

#     # return attr

# # For decoder models, only Layered Integrated Gradients is supported
# def compute_gradients_causal(wrapper, prompt, output):
#     # LayerIntegratedGradients is the? only supported for decoder models
#     # TODO: can we improve this forward function? It is taking a lot of time at the moment
#     ig = LayerIntegratedGradients(forward_func=wrapper.model, layer=wrapper.model.get_output_embeddings())

#     # LLM attribution
#     llm_attr = LLMGradientAttribution(ig, wrapper.tokenizer)
    
#     # TextTokenInput is the only supported input for LLMGradientAttribution
#     input = TextTokenInput(
#         prompt, 
#         wrapper.tokenizer,
#         skip_tokens=[1],  # skip the special token for the start of the text <s>
#     )
    
#     res = llm_attr.attribute(input, target=output)
#     return res.seq_attr.detach().cpu().numpy()





from typing import Tuple
import transformers
import torch
from operator import attrgetter
import time

from helmet.model.base_lm import Base_LM
from helmet.utils.types import *
from helmet.utils.constants import ALTERNATIVES, SALIENCY, CONTRASTIVE
from helmet.explainers.gradients import analyze_token, input_x_gradient

class DEC_LM(Base_LM):
    def __init__(self, model_checkpoint: str, model: transformers.AutoModelForCausalLM, 
                 tokenizer: transformers.PreTrainedTokenizer, url: str, project_id: str, model_config: dict = {}):
        self.model_type = "dec"
        self.model_config = model_config

        try:
            assert "embeddings" in model_config, AssertionError("embeddings must be specified in model_config")
            retriever = attrgetter(model_config["embeddings"])
            embeddings = retriever(model)
            assert embeddings is not None, AssertionError(f"embeddings {model_config['embeddings']} not found in model")
        except Exception as e:
            print(e)
            raise KeyError("embeddings must be specified in model_config")

        super().__init__(model_checkpoint, model, tokenizer, self.model_type, url, project_id, embeddings)
    
    def forward(self, inputs, max_new_tokens, **kwargs) -> Tuple[list, AlternativesExplanation]:
        input_len = len(inputs["input_ids"][0])
        amount_potentials = 5
        
        device = torch.device("cpu")
        self.model.to(device)
        inputs["input_ids"] = inputs["input_ids"].to(device)
        inputs["attention_mask"] = inputs["attention_mask"].to(device)

        output = self.model.generate(
            input_ids=inputs["input_ids"], 
            attention_mask=inputs["attention_mask"],
            use_cache=True, 
            max_new_tokens=max_new_tokens,
            return_dict_in_generate=True, 
            output_scores=True # this gets the scores, while logits are unprocessed.
        )
        alternatives_per_token = []
        for i in range(len(output.scores)):
            scores = output.scores[i]
            top_k = scores.topk(amount_potentials, dim=1)
            top_k_scores = top_k.values.detach().flatten().tolist()
            top_k_indices = top_k.indices

            tokens = self.tokenizer.convert_ids_to_tokens(top_k_indices.detach().flatten(), skip_special_tokens=True)
            res = [{"token": token, "score": score} for token, score in zip(tokens, top_k_scores)]
            alternatives_per_token.append(res)
        
        output_token_ids = output.sequences[0][input_len:]
        return output_token_ids, AlternativesExplanation(alternatives_per_token)


    def explain(self, input, output_token_ids, type: str = "gradient") -> Optional[Explanation]:
        pass
        # if type == "perturbation":
        #     calculate_feature_ablation(self.model, self.tokenizer, prompt, output)
        # return compute_gradients_causal(self, prompt, output)
    
    def saliency_explainer(self, id: str, **kwargs) -> SaliencyExplanation:
        run: Run = self.get_run(id)
        input = self._encode_text(run.input.prompt)
        output_token_ids = self.tokenizer.convert_tokens_to_ids(run.output.tokens)
        
        input_ids = input["input_ids"][0]
        attention_mask = input["attention_mask"]

        merged = torch.cat((input_ids, torch.tensor(output_token_ids)), 0)
        start_index = len(input_ids)
        total_length = len(merged)

        result = []
        for idx in range(start_index, total_length):
            curr_input_ids = merged[:idx]
            output_id = merged[idx]
            base_saliency_matrix, base_embd_matrix = analyze_token(self, curr_input_ids, attention_mask, correct=output_id)
            gradients = input_x_gradient(base_saliency_matrix, base_embd_matrix, normalize=True)
            result.append(gradients)
            print("finished token", idx, "of", total_length)

        explanation = SaliencyExplanation(input_attributions=result)
        run.explanations.append(explanation)

        self.update_run(run)

        return explanation
    
    def contrastive_explainer(self, id: str, alternative_str: str, **kwargs) -> ContrastiveExplanation:
        run: Run = self.get_run(id)
        input = self._tokenize(run.input.prompt)
        alternative_output = self._encode_text(alternative_str)
        output_token_ids = self.tokenizer.convert_tokens_to_ids(run.output.tokens)

        input_ids = input["input_ids"][0]
        attention_mask = input["attention_mask"]
        
        output_id = output_token_ids[0]

        alternative_id = alternative_output["input_ids"][0][0] # not sure why we need this
        saliency_matrix, base_embd_matrix = analyze_token(self, input_ids, attention_mask, correct=output_id, foil=alternative_id)
        gradients = input_x_gradient(saliency_matrix, base_embd_matrix, normalize=True)

        alternative_output_str = self.tokenizer.decode(alternative_id, skip_special_tokens=True)
        
        explanation = ContrastiveExplanation(contrastive_input=alternative_output_str, attributions=gradients)
        run.explanations.append(explanation)

        self.update_run(run)

        return explanation

    def predict(self, prompt, generate_explanations=False, groundtruth=None, *args, **kwargs):
        start = time.time()
        eos_token = False
        max_tokens = kwargs.get("max_new_tokens", 10)
        input = self._encode_text(prompt)
        output_token_ids, alternatives = self.forward(input, max_new_tokens=max_tokens)
        output_str: str = self.token_ids_to_string(output_token_ids)

        # if generate_explanations:
        #     explanation_type = kwargs.get("explanation_type", None)
        #     assert explanation_type is not None, AssertionError("explanation_type must be specified if generate_explanations=True")
            
        #     explanation: Explanation = self.explain(input, output_token_ids, explanation_type)
        #     explanations.append(explanation)
        print("Predicted output:", output_str)
        end = time.time()
        execution_time = end - start

        formatted_run = self._format_run(prompt, output_str, [alternatives], execution_time, groundtruth=groundtruth)

        self.update_run(formatted_run)

        return output_str




    # def explain_from_run(self, id: str, explanation_type="gradient", **kwargs):
    #     start = time.time()
    #     run = self.get_run(id)
    #     explanation = self.explain(run.input.prompt, run.output, explanation_type)
    #     formatted_expl = self._format_explanation(explanation, explanation_type)

    #      # record end time
    #     end = time.time()
    #     execution_time = end - start

    #     formatted_run = self._format_run(prompt, result, alternatives, formatted_expl, execution_time_in_sec=execution_time)
    #     self.update_run(formatted_run)
    #     print("Explanation generated and saved successfully!")


import transformers
from transformers import AutoTokenizer

import torch
from helmet.model import DEC_LM
from helmet.updater import get_or_create_project

url = "http://localhost:4000"

# Mapping from model_type to model class
model_type_to_class = {
    "enc": transformers.AutoModelForSequenceClassification,
    "dec": transformers.AutoModelForCausalLM,
    "enc-dec": transformers.AutoModelForSeq2SeqLM,
}

model_type_to_implementation = {
    # "enc": ENC_LM,
    # "enc-dec": ENC_DEC_LM,
    "dec": DEC_LM,
}

project_setup = [
    "project_id",
    "platform_url",
]

model_setup_args = [
    "checkpoint",
    "model_type",
    "embeddings",
]

run_config_args = [
    "device"
]

def from_pretrained(project_setup: dict = {}, model_setup:dict = {}, run_config: dict={}): 
    platform_url = project_setup.get("platform_url", url)
    project_id = project_setup.get("project_id", None)
    
    model_type = model_setup.get("model_type", None)
    model_checkpoint = model_setup.get("checkpoint", None)

    device = run_config.get("device", "cpu")

    assert device in ["cpu", "auto", "cuda:0"], AssertionError("device must be either 'cpu', 'auto', or 'cuda:0'")
    if device == "cuda:0":
        torch.device(device)
        assert torch.cuda.is_available(), AssertionError("cuda is not available")

    assert model_type in ["enc", "dec", "enc-dec"], AssertionError("model_type must be either 'enc', 'dec', or 'enc-dec'")
    assert project_id is not None, AssertionError("project_id must be specified")

    print("updates will be sent to", platform_url)
    print("setting up model with config", model_setup)

    model_cls = model_type_to_class[model_type]

    hfModel = model_cls.from_pretrained(model_checkpoint, trust_remote_code=True, device_map="auto")
    hfTokenizer = AutoTokenizer.from_pretrained(model_checkpoint)

    modelHelper = model_type_to_implementation[model_type]
    assert modelHelper is not None, AssertionError(f"model_type {model_type} not implemented")

    model = modelHelper(model_checkpoint, hfModel, hfTokenizer, platform_url, project_id, model_setup)
    return model


__all__ = ["from_pretrained", "get_or_create_project"]